
# THIS FILE IS GENERATED FROM PseudoNetCDF SETUP.PY
short_version = '3.2.0'
version = '3.2.0'
full_version = '3.2.0.dev0+c990327'
git_revision = 'c990327ecc68f975848e3b219602d10bf6366ac4'
release = False
if not release:
    version = full_version
