__doc__ = """
.. _racmfiles
:mod:`racmfile` -- RACM File Interfaces
===================================

.. module:: RACM
   :platform: Unix, Windows
   :synopsis: Provides :ref:`PseudoNetCDF`  random access read
              based file interfaces for all RACM related files.  See
              PseudoNetCDF.sci_var.PseudoNetCDFFile for interface details.
.. moduleauthor:: Barron Henderson <barronh@unc.edu>
"""
__all__ = ['boxmodel']

from . import boxmodel
